# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import braintree

import django
from django.conf import settings


settings.configure(
    DEBUG=True,
    DATABASES={
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    PAYMENT_METHOD_SECRET=b'MOW_x1k-ayes3KqnFHNZUxvKipC8iLjxiczEN76TIEA=',
    PAYMENT_PROCESSORS={
        'BraintreeTriggered': {
            'setup_data': {
                'environment': braintree.Environment.Sandbox,
                'merchant_id': "your-merchand-id-here",
                'public_key': "your-public-id-here",
                'private_key': "your-private-id-here"
            },
            'class': 'silver_braintree.payment_processors.BraintreeTriggered',
        },
        'BraintreeTriggeredRecurring': {
            'setup_data': {
                'environment': braintree.Environment.Sandbox,
                'merchant_id': "your-merchand-id-here",
                'public_key': "your-public-id-here",
                'private_key': "your-private-id-here"
            },
            'class': 'silver_braintree.payment_processors.BraintreeTriggeredRecurring'
        },
        'Manual': {
            'class': 'silver.models.payment_processors.manual.ManualProcessor'
        }
    },
    INSTALLED_APPS=(
                    'dal',
                    'dal_select2',
                    'django.contrib.auth',
                    'django.contrib.contenttypes',
                    'django.contrib.sessions',
                    'django.contrib.staticfiles',
                    'django.contrib.admin',
                    'silver',
                    'silver_braintree',),
    CACHES={
        'default': {
            'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
            'LOCATION': 'unique-snowflake',
        }
    },
    USE_TZ=True,
    STATIC_URL='/static/'
)

django.setup()



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from mock import patch, MagicMock
from braintree import Transaction as BraintreeTransaction

from silver.models import Transaction
from silver.payment_processors import get_instance
from silver_braintree.models.customer_data import CustomerData
from silver_braintree.payment_processors import (BraintreeTriggered,
                                                 BraintreeTriggeredRecurring)
from tests.factories import BraintreeTransactionFactory


class TestBraintreeTransactions:
    def setup_method(self):
        BraintreeTriggered._has_been_setup = True
        BraintreeTriggeredRecurring._has_been_setup = True

        transaction = MagicMock()
        transaction.amount = 1000
        transaction.status = BraintreeTransaction.Status.Settled
        transaction.id = 'beertrain'
        transaction.processor_response_code = 2000
        transaction.payment_instrument_type = 'paypal_account'

        transaction.paypal_details = MagicMock()
        transaction.paypal_details.image_url = 'image_url'
        transaction.paypal_details.payer_email = 'payer_email'
        transaction.paypal_details.token = 'kento'

        transaction.credit_card_details = MagicMock()
        transaction.credit_card_details.image_url = 'image_url'
        transaction.credit_card_details.last_4 = '1234'
        transaction.credit_card_details.card_type = 'card_type'
        transaction.credit_card_details.token = 'kento'

        transaction.customer_details = MagicMock()
        transaction.customer_details.id = 'braintree_id'

        self.transaction = transaction

        result = MagicMock()
        result.is_success = True
        result.transaction = transaction
        result.errors = MagicMock()
        result.errors.deep_errors = [MagicMock(code=2001), MagicMock(code=2042)]

        self.result = result

    def teardown_method(self):
        BraintreeTriggered._has_been_setup = False
        BraintreeTriggeredRecurring._has_been_setup = False

    @pytest.mark.django_db
    def test_update_status_transaction_settle(self):
        transaction = BraintreeTransactionFactory.create(
            state=Transaction.States.Pending, data={
                'braintree_id': 'beertrain'
            }
        )

        with patch('braintree.Transaction.find') as find_mock:
            find_mock.return_value = self.transaction
            payment_processor = get_instance(transaction.payment_processor)
            payment_processor.fetch_transaction_status(transaction)

            find_mock.assert_called_once_with('beertrain')

            assert transaction.state == transaction.States.Settled
            assert transaction.data.get('status') == self.result.transaction.status

    @pytest.mark.django_db
    def test_update_status_transaction_fail(self):
        transaction = BraintreeTransactionFactory.create(
            state=Transaction.States.Pending, data={
                'braintree_id': 'beertrain'
            }
        )
        with patch('braintree.Transaction.find') as find_mock:
            find_mock.return_value = self.transaction
            # fail status from braintree
            self.transaction.status = BraintreeTransaction.Status.ProcessorDeclined

            payment_processor = get_instance(transaction.payment_processor)
            payment_processor.fetch_transaction_status(transaction)

            find_mock.assert_called_once_with('beertrain')

            assert transaction.state == transaction.States.Failed
            assert transaction.data.get('status') == self.result.transaction.status

    @pytest.mark.django_db
    def test_execute_transaction_with_nonce_nonrecurring(self):
        transaction = BraintreeTransactionFactory.create()
        payment_method = transaction.payment_method
        payment_method.nonce = 'some-nonce'
        payment_method.save()

        with patch('braintree.Transaction.sale') as sale_mock:
            sale_mock.return_value = self.result
            payment_processor = get_instance(transaction.payment_processor)
            payment_processor.execute_transaction(transaction)

            sale_mock.assert_called_once_with({
                # new customer
                'customer': {'first_name': payment_method.customer.first_name,
                             'last_name': payment_method.customer.last_name},
                'amount': transaction.amount,
                'billing': {'postal_code': None},
                # don't store payment method in vault
                'options': {'store_in_vault': False,
                            'submit_for_settlement': True},
                'payment_method_nonce': payment_method.nonce
            })

            assert transaction.state == transaction.States.Settled

            payment_method.refresh_from_db()
            assert not payment_method.token

            assert payment_method.details == {
                'image_url': self.transaction.paypal_details.image_url,
                'email': self.transaction.paypal_details.payer_email,
                'type': self.transaction.payment_instrument_type,
            }

            assert (payment_method.display_info ==
                    self.transaction.paypal_details.payer_email)

            customer = transaction.customer
            customer_data = CustomerData.objects.filter(customer=customer)
            assert len(customer_data) == 1

            customer_data = customer_data[0]
            assert customer_data.get('id') == self.transaction.customer_details.id

            assert transaction.data.get('status') == self.result.transaction.status

    @pytest.mark.django_db
    def test_execute_transaction_with_token_recurring(self):
        transaction = BraintreeTransactionFactory.create()
        transaction.payment_method.payment_processor = 'BraintreeTriggeredRecurring'
        transaction.payment_method.save()

        payment_method = transaction.payment_method
        payment_method.token = 'kento'
        payment_method.save()

        customer = payment_method.customer
        customer_data = CustomerData.objects.create(
            customer=customer, data={'id': 'somethingelse'}
        )

        with patch('braintree.Transaction.sale') as sale_mock:
            sale_mock.return_value = self.result

            payment_processor = get_instance(transaction.payment_processor)
            payment_processor.execute_transaction(transaction)

            sale_mock.assert_called_once_with({
                # existing customer in vault
                'customer_id': customer_data['id'],
                'amount': transaction.amount,
                'billing': {'postal_code': None},
                'options': {'submit_for_settlement': True},
                # existing token
                'payment_method_token': payment_method.token
            })

            assert transaction.state == transaction.States.Settled

            payment_method.refresh_from_db()
            assert payment_method.token == self.transaction.paypal_details.token
            assert payment_method.details == {
                'image_url': self.transaction.paypal_details.image_url,
                'email': self.transaction.paypal_details.payer_email,
                'type': self.transaction.payment_instrument_type,
            }
            assert payment_method.verified
            assert (payment_method.display_info ==
                    self.transaction.paypal_details.payer_email)

            assert transaction.data.get('status') == self.result.transaction.status

    @pytest.mark.django_db
    def test_execute_transaction_with_nonce_recurring_paypal(self):
        transaction = BraintreeTransactionFactory.create()
        transaction.payment_method.payment_processor = 'BraintreeTriggeredRecurring'
        transaction.payment_method.save()

        nonce = 'some-nonce'
        payment_method = transaction.payment_method
        payment_method.nonce = nonce
        payment_method.save()

        with patch('braintree.Transaction.sale') as sale_mock:
            sale_mock.return_value = self.result

            payment_processor = get_instance(transaction.payment_processor)
            payment_processor.execute_transaction(transaction)

            sale_mock.assert_called_once_with({
                # create new customer in vault
                'customer': {'first_name': payment_method.customer.first_name,
                             'last_name': payment_method.customer.last_name},
                'amount': transaction.amount,
                'billing': {'postal_code': None},
                # store the payment method
                'options': {'store_in_vault': True,
                            'submit_for_settlement': True},
                'payment_method_nonce': nonce
            })

            assert transaction.state == transaction.States.Settled

            payment_method.refresh_from_db()
            assert payment_method.token == self.transaction.paypal_details.token
            assert payment_method.details == {
                'image_url': self.transaction.paypal_details.image_url,
                'email': self.transaction.paypal_details.payer_email,
                'type': self.transaction.payment_instrument_type,
            }
            assert payment_method.verified
            assert not payment_method.nonce
            assert (payment_method.display_info ==
                    self.transaction.paypal_details.payer_email)

            assert transaction.data.get('status') == self.result.transaction.status

            customer = transaction.customer

            customer_data = CustomerData.objects.filter(customer=customer)
            assert len(customer_data) == 1

            customer_data = customer_data[0]
            assert customer_data.get('id') == self.transaction.customer_details.id

    @pytest.mark.django_db
    def test_execute_transaction_with_nonce_recurring_credit_card(self):
        transaction = BraintreeTransactionFactory.create()
        transaction.payment_method.payment_processor = 'BraintreeTriggeredRecurring'
        transaction.payment_method.save()

        nonce = 'some-nonce'
        payment_method = transaction.payment_method
        payment_method.nonce = nonce
        payment_method.save()

        with patch('braintree.Transaction.sale') as sale_mock:
            self.result.transaction.payment_instrument_type = 'credit_card'
            sale_mock.return_value = self.result

            payment_processor = get_instance(transaction.payment_processor)
            payment_processor.execute_transaction(transaction)

            sale_mock.assert_called_once_with({
                # create new customer in vault
                'customer': {'first_name': payment_method.customer.first_name,
                             'last_name': payment_method.customer.last_name},
                'amount': transaction.amount,
                'billing': {'postal_code': None},
                # store the payment method
                'options': {'store_in_vault': True,
                            'submit_for_settlement': True},
                'payment_method_nonce': nonce
            })

            assert transaction.state == transaction.States.Settled

            payment_method.refresh_from_db()
            assert payment_method.token == self.transaction.credit_card_details.token
            assert payment_method.details == {
                'last_4': self.transaction.credit_card_details.last_4,
                'card_type': self.transaction.credit_card_details.card_type,
                'type': self.transaction.payment_instrument_type,
                'image_url': self.transaction.credit_card_details.image_url
            }
            assert payment_method.verified
            assert not payment_method.nonce

            assert transaction.data.get('status') == self.result.transaction.status

            customer = transaction.customer

            customer_data = CustomerData.objects.filter(customer=customer)
            assert len(customer_data) == 1

            customer_data = customer_data[0]

            assert customer_data.get('id') == self.transaction.customer_details.id

    @pytest.mark.django_db
    def test_execute_transaction_with_canceled_payment_method(self):
        transaction = BraintreeTransactionFactory.create(
            payment_processor='BraintreeTriggered'
        )

        nonce = 'some-nonce'
        payment_method = transaction.payment_method
        payment_method.nonce = nonce
        payment_method.canceled = True
        payment_method.save()

        with patch('braintree.Transaction.sale') as sale_mock:
            payment_processor = get_instance(transaction.payment_processor)
            assert payment_processor.execute_transaction(transaction) is False
            assert sale_mock.call_count == 0

    @pytest.mark.django_db
    def test_execute_transaction_with_payment_method_without_nonce_or_token(self):
        transaction = BraintreeTransactionFactory.create(
            payment_processor='BraintreeTriggered'
        )

        with patch('braintree.Transaction.sale') as sale_mock:
            payment_processor = get_instance(transaction.payment_processor)

            assert payment_processor.execute_transaction(transaction) is False
            assert sale_mock.call_count == 0

    @pytest.mark.django_db
    def test_execute_transaction_failed_braintree_response(self):
        transaction = BraintreeTransactionFactory.create(
            payment_processor='BraintreeTriggered'
        )

        nonce = 'some-nonce'
        payment_method = transaction.payment_method
        payment_method.nonce = nonce
        payment_method.save()

        with patch('braintree.Transaction.sale') as sale_mock:
            self.result.is_success = False
            sale_mock.return_value = self.result

            payment_processor = get_instance(transaction.payment_processor)
            assert payment_processor.execute_transaction(transaction) is False
            assert transaction.data.get('error_codes') == [
                error.code for error in self.result.errors.deep_errors
            ]



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import braintree
from braintree.exceptions import (AuthenticationError, AuthorizationError,
                                  DownForMaintenanceError, ServerError,
                                  UpgradeRequiredError)
from django_fsm import TransitionNotAllowed


from silver.payment_processors import PaymentProcessorBase, get_instance
from silver.payment_processors.forms import GenericTransactionForm
from silver.payment_processors.mixins import TriggeredProcessorMixin

from silver_braintree.models import BraintreePaymentMethod
from silver_braintree.models.customer_data import CustomerData
from silver_braintree.views import BraintreeTransactionView


logger = logging.getLogger(__name__)


class BraintreeTriggeredBase(PaymentProcessorBase, TriggeredProcessorMixin):
    payment_method_class = BraintreePaymentMethod
    transaction_view_class = BraintreeTransactionView
    form_class = GenericTransactionForm
    template_slug = 'braintree'

    _has_been_setup = False

    def is_payment_method_recurring(self, payment_method):
        raise NotImplementedError

    def __init__(self, name, *args, **kwargs):
        super(BraintreeTriggeredBase, self).__init__(name)

        if self._has_been_setup:
            return

        environment = kwargs.pop('environment', None)
        braintree.Configuration.configure(environment, **kwargs)

        BraintreeTriggeredBase._has_been_setup = True

    def client_token(self, customer):
        customer_data = CustomerData.objects.get_or_create(customer=customer)[0]
        customer_braintree_id = customer_data.get('id')

        try:
            return braintree.ClientToken.generate(
                {'customer_id': customer_braintree_id}
            )
        except (AuthenticationError, AuthorizationError, DownForMaintenanceError,
                ServerError, UpgradeRequiredError) as e:
            logger.warning(
                'Couldn\'t obtain Braintree client_token %s', {
                    'customer_id': customer_braintree_id,
                    'exception': str(e)
                }
            )

    def refund_transaction(self, transaction, payment_method=None):
        pass

    def void_transaction(self, transaction, payment_method=None):
        pass

    def _update_payment_method(self, payment_method, result_details,
                               instrument_type):
        """
        :param payment_method: A BraintreePaymentMethod.
        :param result_details: A (part of) braintreeSDK result(response)
                               containing payment method information.
        :param instrument_type: The type of the instrument (payment method);
                                see BraintreePaymentMethod.Types.
        :description: Updates a given payment method's data with data from a
                      braintreeSDK result payment method.
        """
        payment_method_details = {
            'type': instrument_type,
            'image_url': result_details.image_url,
        }

        if instrument_type == payment_method.Types.PayPal:
            payment_method_details['email'] = result_details.payer_email
            payment_method.display_info = result_details.payer_email
        elif instrument_type == payment_method.Types.CreditCard:
            payment_method_details.update({
                'card_type': result_details.card_type,
                'last_4': result_details.last_4,
            })

        payment_method.update_details(payment_method_details)

        if self.is_payment_method_recurring(payment_method):
            if result_details.token:
                payment_method.token = result_details.token
                payment_method.data.pop('nonce', None)
                payment_method.verified = True

        payment_method.save()

    def _update_transaction_status(self, transaction, result_transaction):
        """
        :param transaction: A Transaction.
        :param result_transaction: A transaction from a braintreeSDK
                                   result(response).
        :description: Updates a given transaction's data with data from a
                      braintreeSDK result payment method.
        :returns True if transaction is on the happy path, False otherwise.
        """
        if not transaction.data:
            transaction.data = {}

        transaction.external_reference = result_transaction.id
        status = result_transaction.status

        transaction.data['status'] = status

        target_state = None
        try:
            if status in [braintree.Transaction.Status.AuthorizationExpired,
                          braintree.Transaction.Status.SettlementDeclined,
                          braintree.Transaction.Status.Failed,
                          braintree.Transaction.Status.GatewayRejected,
                          braintree.Transaction.Status.ProcessorDeclined]:
                target_state = transaction.States.Failed
                if transaction.state != target_state:
                    fail_code = self._get_silver_fail_code(result_transaction)
                    fail_reason = self._get_braintree_fail_code(result_transaction)
                    transaction.fail(fail_code=fail_code, fail_reason=fail_reason)
                    return False

            elif status == braintree.Transaction.Status.Voided:
                target_state = transaction.States.Canceled
                if transaction.state != target_state:
                    transaction.cancel()
                    return False

            elif status in [braintree.Transaction.Status.Settling,
                            braintree.Transaction.Status.SettlementPending,
                            braintree.Transaction.Status.Settled]:
                target_state = transaction.States.Settled
                if transaction.state != target_state:
                    transaction.settle()
                    return True
            else:
                return True
        except TransitionNotAllowed as e:
            logger.warning('Braintree Transaction couldn\'t transition locally: '
                           '%s' % {
                               'initial_state': transaction.state,
                               'target_state': target_state,
                               'transaction_id': transaction.id,
                               'transaction_uuid': transaction.uuid
                           })
            raise e
        finally:
            transaction.save()

    def _update_customer(self, customer, result_details):
        customer_data = CustomerData.objects.get_or_create(customer=customer)[0]
        if 'id' not in customer_data:
            customer_data['id'] = result_details.id
            customer_data.save()

    def _get_errors(self, result):
        return [
            error.code for error in result.errors.deep_errors
        ] if result.errors else None

    def _get_braintree_fail_code(self, result_transaction):
        if result_transaction.status in (
                braintree.Transaction.Status.ProcessorDeclined,
                braintree.Transaction.Status.AuthorizationExpired
        ):
            return result_transaction.processor_response_code,

        elif result_transaction.status in (
                braintree.Transaction.Status.SettlementDeclined,
                braintree.Transaction.Status.SettlementFailed
        ):
            return result_transaction.processor_settlement_response_code,

    def _get_silver_fail_code(self, result_transaction):
        braintree_fail_code = self._get_braintree_fail_code(result_transaction)

        if not braintree_fail_code:
            return 'default'

        if braintree_fail_code in [2001]:
            return 'insufficient_funds'
        elif braintree_fail_code in [2022]:
            return 'expired_payment_method'
        elif braintree_fail_code in [2004, 2022]:
            return 'expired_card'
        elif braintree_fail_code in [2071, 2072, 2073]:
            return 'invalid_payment_method'
        elif braintree_fail_code in [2005, 2006]:
            return 'invalid_card'
        elif braintree_fail_code in [2002, 2003, 2086]:
            return 'limit_exceeded'
        elif braintree_fail_code in [2000, 2010, 2019, 2038, 2046, 2061, 2062,
                                     2064]:
            return 'transaction_declined_by_bank'
        elif braintree_fail_code in [2075, 2076, 2077]:
            return 'transaction_hard_declined'
        elif braintree_fail_code in [2012, 2013, 2014, 2074]:
            return 'transaction_hard_declined_by_bank'

        return 'default'

    def _charge_transaction(self, transaction):
        """
        :param transaction: The transaction to be charged. Must have a useable
                            payment_method.
        :return: True on success, False on failure.
        """
        payment_method = transaction.payment_method

        if payment_method.canceled:
            try:
                transaction.fail(fail_reason='Payment method was canceled.')
                transaction.save()
            finally:
                return False

        # prepare payload
        options = {
            'submit_for_settlement': True,
        }

        if payment_method.token:
            payload = {'payment_method_token': payment_method.token}
        elif payment_method.nonce:
            options.update({
                "store_in_vault": self.is_payment_method_recurring(payment_method)
            })
            payload = {'payment_method_nonce': payment_method.nonce}
        else:
            logger.warning('Token or nonce not found when charging '
                           'BraintreePaymentMethod: %s', {
                               'payment_method_id': payment_method.id
                           })

            try:
                transaction.fail(fail_reason='Payment method has no token or nonce.')
                transaction.save()
            finally:
                return False

        payload.update({
            'amount': transaction.amount,
            'billing': {
                'postal_code': payment_method.data.get('postal_code')
            },
            # TODO check how firstname and lastname can be obtained (for both
            # credit card and paypal)
            'options': options
        })

        customer = transaction.customer
        customer_data = CustomerData.objects.get_or_create(customer=customer)[0]

        if 'id' in customer_data:
            payload.update({
                'customer_id': customer_data['id']
            })
        else:
            payload.update({
                'customer': {
                    'first_name': customer.first_name,
                    'last_name': customer.last_name
                }
            })

        # send transaction request
        result = braintree.Transaction.sale(payload)

        # handle response
        if not result.is_success or not result.transaction:
            errors = self._get_errors(result)
            logger.warning('Couldn\'t charge Braintree transaction.: %s', {
                'message': result.message,
                'errors': errors,
                'customer_id': customer.id,
                'card_verification': (result.credit_card_verification if errors
                                      else None)
            })

            transaction.data['error_codes'] = errors
            try:
                fail_code = self._get_silver_fail_code(result.transaction)
                transaction.fail(fail_code=fail_code, fail_reason=errors)
            finally:
                transaction.save()

                return False

        self._update_customer(customer, result.transaction.customer_details)

        instrument_type = result.transaction.payment_instrument_type

        if instrument_type == payment_method.Types.PayPal:
            details = result.transaction.paypal_details
        elif instrument_type == payment_method.Types.CreditCard:
            details = result.transaction.credit_card_details
        else:
            # Only PayPal and CreditCard are currently handled
            try:
                transaction.fail(fail_code='invalid_payment_method',
                                 fail_reason='Not a supported instrument_type')
                transaction.save()
            finally:
                return False

        self._update_payment_method(
            payment_method, details, instrument_type
        )

        try:
            return self._update_transaction_status(transaction,
                                                   result.transaction)
        except TransitionNotAllowed:
            # ToDo handle this
            return False

    def execute_transaction(self, transaction):
        """
        :param transaction: A Braintree transaction in Initial state.
        :return: True on success, False on failure.
        """

        payment_processor = get_instance(transaction.payment_processor)
        if not payment_processor == self:
            return False

        if transaction.state != transaction.States.Initial:
            return False

        try:
            transaction.process()
            transaction.save()
        except TransitionNotAllowed:
            return False

        return self._charge_transaction(transaction)

    def fetch_transaction_status(self, transaction):
        """
        :param transaction: A Braintree transaction in Initial or Pending state.
        :return: True on success, False on failure.
        """

        payment_processor = get_instance(transaction.payment_processor)
        if not payment_processor == self:
            return False

        if transaction.state != transaction.States.Pending:
            return False

        if not transaction.data.get('braintree_id'):
            logger.warning('Found pending Braintree transaction with no '
                           'braintree_id: %s', {
                                'transaction_id': transaction.id,
                                'transaction_uuid': transaction.uuid
                           })

            return False

        try:
            result_transaction = braintree.Transaction.find(
                transaction.data['braintree_id']
            )
            return self._update_transaction_status(transaction,
                                                   result_transaction)
        except braintree.exceptions.NotFoundError:
            logger.warning('Couldn\'t find Braintree transaction from '
                           'Braintree %s', {
                                'braintree_id': transaction.data['braintree_id'],
                                'transaction_id': transaction.id,
                                'transaction_uuid': transaction.uuid
                           })
            return False
        except TransitionNotAllowed:
            return False
            # ToDo handle this

    def handle_transaction_response(self, transaction, request):
        payment_method_nonce = request.POST.get('payment_method_nonce')

        payment_method = transaction.payment_method
        if not payment_method_nonce:
            try:
                transaction.fail(fail_reason='payment_method_nonce was not provided.')
                transaction.save()
            except TransitionNotAllowed:
                pass
            finally:
                return

        if payment_method.nonce:
            try:
                transaction.fail(fail_reason='Payment method already has a nonce.')
                transaction.save()
            except TransitionNotAllowed:
                pass
            finally:
                return

        # initialize the payment method
        details = {
            'postal_code': request.POST.get('postal_code')
        }

        payment_method.nonce = payment_method_nonce
        payment_method.update_details(details)
        payment_method.save()

        # manage the transaction
        payment_processor = get_instance(payment_method.payment_processor)

        if not payment_processor.execute_transaction(transaction):
            try:
                transaction.fail()
                transaction.save()
            except TransitionNotAllowed:
                pass
            finally:
                return


class BraintreeTriggered(BraintreeTriggeredBase):
    def is_payment_method_recurring(self, payment_method):
        return False


class BraintreeTriggeredRecurring(BraintreeTriggeredBase):
    def is_payment_method_recurring(self, payment_method):
        return True



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import jsonfield
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('silver', '0032_auto_20170201_1342'),
    ]

    operations = [
        migrations.CreateModel(
            name='BraintreePaymentMethod',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('silver.paymentmethod',),
        ),
        migrations.CreateModel(
            name='CustomerData',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('customer', models.ForeignKey(to='silver.Customer')),
            ],
        ),
    ]



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import braintree as sdk
from braintree.exceptions import NotFoundError

from silver.models import PaymentMethod


class BraintreePaymentMethod(PaymentMethod):
    """
        data field structure
        {
            'nonce': 'some-nonce', (encrypted, deleted if token exists)
            'token': 'some-token', (encrypted)
            'braintree_id': 'transaction-id-given-by-braintree',
            'status': 'status-given-by-braintree' (does not exist if
                                                   Transaction.state is Initial)
            'details': {
                - common -
                'type': Types.PayPal | Types.CreditCard,
                'image_url': 'payment-processor-icon-url-given-by-braintree',
                - PayPal -
                'email': 'some@ema.il' (PayPal account email)
                - CreditCard -
                'last_4': '1234' (last 4 digits from the credit card number)
                'card_type': e.g. 'Visa',
                'postal_code': '41234Y' (if provided from template)
            }
        }
    """
    class Meta:
        proxy = True

    class Types:
        PayPal = 'paypal_account'
        CreditCard = 'credit_card'

    @property
    def braintree_transaction(self):
        try:
            return sdk.Transaction.find(self.braintree_id)
        except NotFoundError:
            return None

    @property
    def braintree_id(self):
        return self.data.get('braintree_id')

    @property
    def token(self):
        return self.decrypt_data(self.data.get('token'))

    @token.setter
    def token(self, value):
        self.data['token'] = self.encrypt_data(value)

    @property
    def nonce(self):
        return self.decrypt_data(self.data.get('nonce'))

    @nonce.setter
    def nonce(self, value):
        self.data['nonce'] = self.encrypt_data(value)

    def update_details(self, details):
        if 'details' not in self.data:
            self.data['details'] = details
        else:
            self.data['details'].update(details)

    @property
    def details(self):
        return self.data.get('details')

    @property
    def public_data(self):
        return self.data.get('details')



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from silver.payment_processors import get_instance
from silver.payment_processors.views import GenericTransactionView


class BraintreeTransactionView(GenericTransactionView):
    def get_context_data(self):
        context_data = super(BraintreeTransactionView, self).get_context_data()
        payment_processor = get_instance(self.transaction.payment_processor)
        context_data['client_token'] = payment_processor.client_token(
                self.transaction.customer
        )
        context_data['is_recurring'] = payment_processor.is_payment_method_recurring(
            self.transaction.payment_method
        )

        return context_data



from jsonfield import JSONField
from silver.models import Customer

from django.core.exceptions import ValidationError
from django.db.models import Model, ForeignKey


class CustomerData(Model):
    customer = ForeignKey(Customer)
    data = JSONField(default={}, null=True, blank=True)

    def clean(self):
        super(CustomerData, self).clean()

        if not isinstance(self.data, dict):
            raise ValidationError('Field "data" must be a dict.')

    def __repr__(self):
        return '%s Braintree data' % self.customer

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, item):
        return item in self.data

    def get(self, key, default=None):
        return self.data.get(key, default)

    def update(self, data):
        return self.data.update(data)

    def pop(self, key, default=None):
        return self.data.pop(key, default)



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from annoying.functions import get_object_or_None

from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response

from silver.models import Transaction
from silver.payment_processors import get_instance

from silver_braintree.models import BraintreeTriggered


@api_view('GET')
def client_token(request, transaction_uuid=None):
    transaction = get_object_or_None(Transaction, id=transaction_uuid)

    payment_processor = get_instance(transaction.payment_processor)
    if not isinstance(payment_processor, BraintreeTriggered):
        return Response(
            {'detail': 'Transaction is not a Braintree transaction.'},
            status=status.HTTP_400_BAD_REQUEST
        )

    token = payment_processor.client_token(transaction.customer)

    if not token:
        return Response({'detail': 'Braintree miscommunication.'},
                        status=status.HTTP_503_SERVICE_UNAVAILABLE)

    return Response({'token': token}, status=status.HTTP_200_OK)



# Copyright (c) 2017 Presslabs SRL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -*- encoding: utf-8 -*-

import os

import versioneer
from setuptools import setup, find_packages

from silver_braintree import __version__ as version

install_requires = ['braintree']


def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except IOError:
        return ''

setup(
    name="silver-braintree",
    version=version,
    cmdclass=versioneer.get_cmdclass(),
    description=read('DESCRIPTION'),
    long_description=read('README.md'),
    license='Apache 2.0',
    platforms=['OS Independent'],
    keywords='django, app, reusable, billing, invoicing, api',
    author='Presslabs',
    author_email='ping@presslabs.com',
    url='https://github.com/silverapp/silver-braintree',
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django :: 1.8',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7'
    ]
)
